/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <netndfsm.h>
#include <atl/arraylist.h>
#include <atl/integer.h>
#include <atl/int2ptrhash.h>
#include <atl/hashmap.h>
#include <utils/assertion.h>

// Node weighing is turned on
#define NDFSM_NODE_WEIGHING

// Nondeterministic arcs grouping is turned on 
#define NDFSM_ARC_GROUPING

#ifdef NDFSM_NODE_WEIGHING
    #include <atl/pqueue.h>
#endif

/******************************************************************************/
/**                  NDFSM Internal Parameters and Macroses                  **/
/******************************************************************************/

#define NDFSM_NODES_ALLOC_BY            16536
#define NDFSM_NODES_HASH_SIZE           131072
#define NDFSM_PATH_FINDER_MAX_HASH_SIZE 131072
#define NDFSM_PATH_FINDER_MIN_HASH_SIZE 1024
#define NDFSM_PATH_ALLOC_BY             1024
#define NDFSM_ARCS_ALLOC_BY             32

#define NDFSM_OPTIMISTIC_WEIGHING

#ifdef NDFSM_NODE_WEIGHING
    #define NDFSM_INITIAL_WEIGHT(node) 0
 #ifdef NDFSM_OPTIMISTIC_WEIGHING
    #define NDFSM_UNKNOWN_WEIGHT(data) 1
 #else
    #define NDFSM_UNKNOWN_WEIGHT(data) ((int)((float)(data)->path_sum/(data)->path_num))
 #endif
    #define NDFSM_MAXIMUM_WEIGHT       0x3fffffff
    #define NDFSM_MIN_WEIGHT(lhs, rhs) ((lhs) < (rhs) ? (lhs) : (rhs))
    #define NDFSM_MAX_WEIGHT(lhs, rhs) ((lhs) > (rhs) ? (lhs) : (rhs))
    #define NDFSM_AVG_PATH_SAMPLE      100
#endif

#ifdef NDFSM_ARC_GROUPING
    #define NDFSM_MAX_ARC_GROUP_SIZE   8
    #define NDFSM_END_NODES_ALLOC_BY   1
#endif

#define NODE_ID(node) \
    ((node) != NULL ? (node)->node_id : NULL)

#define NODE_STRING(node) \
    ((node) != NULL ? NODE_ID_STRING((node)->node_id) : "unknown")

#define GET_START_NODE(arc) \
    ((arc)->start_node)

#ifdef NDFSM_ARC_GROUPING
    #define GET_END_NODE(arc) \
        ((arc)->end_nodes.size > 0 ? *((NDFSM_Node**)get_ArrayList(&(arc)->end_nodes, 0)) : NULL)
    #define GET_END_NODE_I(arc, i) \
        (*((NDFSM_Node**)get_ArrayList(&(arc)->end_nodes, i)))
#else
    #define GET_END_NODE(arc) \
        ((arc)->end_node)
#endif

/******************************************************************************/
/**                             NDFSM Error Codes                            **/
/******************************************************************************/

typedef enum NDFSM_Errno
{
    NDFSM_ERRNO_NO_ERRORS = 0x0,
    NDFSM_ERRNO_UNCONNECT = 0x1,
} NDFSM_Errno;

/******************************************************************************/
/**                             NDFSM Node and Arc                           **/
/******************************************************************************/

/** @typedef NDFSM_Node
 * Internal representation of a graph node.
 */
typedef struct NDFSM_Node
{
    /// Internal node identifier (index in a list of nodes).
    int index;
    /// External node identifier.
    TE_NodeId *node_id;
    /// Untraversed out-going arcs (end nodes of untraversed arcs are null).
    ArrayList new_arcs;
    /// Traversed arcs (including nondeterministic ones).
    ArrayList old_arcs;
#ifdef NDFSM_NODE_WEIGHING
    int weight;
    int distance;
#endif
} NDFSM_Node;

/** @typedef NDFSM_ArcFlag
 * Flags representing arc properties.
 */
typedef enum NDFSM_ArcFlag
{
    /// Nondeterministic stimulus.
    NDFSM_FLAG_NDETERM = 0x1,
    /// Chaotic stimulus (too much nondeterministic).
    NDFSM_FLAG_CHAOTIC = 0x2
} NDFSM_ArcFlags;

/** @typedef NDFSM_Arc
 * Internal representation of a graph arc.
 */
typedef struct NDFSM_Arc
{
    /// Internal arc identifier (order number).
    int index;
    /// Arc flags (see NDFSM_ArcFlags).
    int flags;
    /// Stimulus identifier.
    TE_StimulusId stimulus_id;
    /// Start node of the arc.
    NDFSM_Node *start_node;
#ifdef NDFSM_ARC_GROUPING
    /// End nodes of the arc.
    ArrayList end_nodes;
    /// Number of nondeterministic degree of a path.
    int ndegree;
#else
    /// End node of the arc.
    NDFSM_Node *end_node;
#endif
} NDFSM_Arc;

static void init_NDFSM_Node(NDFSM_Node *node, int index, TE_NodeId *node_id)
{
    assertion(node != NULL, "init_NDFSM_Node: node == NULL");

    node->index = index;
    node->node_id = r(node_id);

    init_ArrayList(&node->new_arcs, sizeof(NDFSM_Arc), NDFSM_ARCS_ALLOC_BY, NULL);
    init_ArrayList(&node->old_arcs, sizeof(NDFSM_Arc), NDFSM_ARCS_ALLOC_BY, NULL);

#ifdef NDFSM_NODE_WEIGHING
    node->weight    = 0;
    node->distance  = 0;
#endif
}

#ifdef NDFSM_NODE_WEIGHING
static void setWeight_NDFSM_Node(NDFSM_Node *node)
{
    int i, w = NDFSM_MAXIMUM_WEIGHT;

    if(node->new_arcs.size > 0)
        { node->weight = 0; return; }

    for(i = 0; i < node->old_arcs.size; i++)
    {
        NDFSM_Arc *arc = (NDFSM_Arc*)get_ArrayList(&node->old_arcs, i);
        NDFSM_Node *end_node = GET_END_NODE(arc);

        if((arc->flags & NDFSM_FLAG_NDETERM) == 0 && end_node->weight < w)
            { w = end_node->weight; }
    }

    node->weight = w + 1;
}
#endif

static void destroy_NDFSM_Node(NDFSM_Node *node)
{
    assertion(node != NULL, "destroy_NDFSM_Node: node == NULL");

    destroy(node->node_id);

    // If the node is finished, then the list of untraversed arcs is destroyed
    if(node->new_arcs.size > 0)
        { destroy_ArrayList(&node->new_arcs); }

    destroy_ArrayList(&node->old_arcs);
}

static void init_NDFSM_Arc(NDFSM_Arc *arc, int index, int flags,
    TE_StimulusId stimulus_id, NDFSM_Node *start_node, NDFSM_Node *end_node)
{
    assertion(arc != NULL, "init_NDFSM_Arc: arc == NULL");

    arc->index       = index;
    arc->flags       = flags;
    arc->stimulus_id = stimulus_id;
    arc->start_node  = start_node;

#ifdef NDFSM_ARC_GROUPING
    init_ArrayList(&arc->end_nodes, sizeof(NDFSM_Node*), NDFSM_END_NODES_ALLOC_BY, NULL);
    arc->ndegree = 0;

    // Append the end node into the list
    if(end_node != NULL)
    {
        NDFSM_Node **node_holder = (NDFSM_Node**)append_ArrayList(&arc->end_nodes);
       *node_holder = end_node;
    }
#else
    arc->end_node = end_node;
#endif
}

static TE_ArcId* createArcId_NDFSM_Arc(NDFSM_Arc *arc)
{
    TE_NodeId *start_node_id = NODE_ID(GET_START_NODE(arc));
    TE_NodeId *end_node_id   = NODE_ID(GET_END_NODE(arc));

    // If the arc is nondeterministic, the first end node is used
    return create_ArcId(r(start_node_id), arc->stimulus_id, r(end_node_id));
}

/******************************************************************************/
/**                         NDFSM Path Finder Data                           **/
/******************************************************************************/

#define NULL_NDFSM_ARC ((NDFSM_Arc*)-1)

typedef bool (*NodeFinder)(NDFSM_Node *node, NDFSM_Node *target_node);

typedef struct PathFinderData
{
    /// Target node (it might be NULL).
    NDFSM_Node *target_node;
    /// Node finder.
    NodeFinder node_finder;
    /// Nodes having been explored.
    ArrayList nodes;
    /// Maps a node index into the arc that leads to the node.
    Int2PtrHashMap nodes_map;
    /// Flag that indicates whether nondeterminism is supported.
    int nondeterm;
#ifdef NDFSM_NODE_WEIGHING
    PQueue *nodes_queue;
#endif
} PathFinderData;

static void init_PathFinderData(PathFinderData *path_data, NDFSM_Node *from, NDFSM_Node *to,
    NodeFinder node_finder, int nondeterm, int hash_size, int min_hash_size, int init_type)
{
    NDFSM_Node **node_holder;
    bool find_nonfinished_node = (to == NULL);

    // Initialize the target node
    path_data->target_node = to;
    // Initialize the node finder
    path_data->node_finder = node_finder;
    // Support of nondeterministic graphs
    path_data->nondeterm = nondeterm;

    // Initialize the nodes list
    init_ArrayList(&path_data->nodes, sizeof(NDFSM_Node*), NDFSM_NODES_ALLOC_BY, NULL);

    // Append the current node into the nodes list
    node_holder = (NDFSM_Node**)append_ArrayList(&path_data->nodes);
   *node_holder = from;

    // Initialize the hash map
    if(init_type)
        { init_Int2PtrHashMap(&path_data->nodes_map, hash_size, min_hash_size); }

    // Append the pair (from -> null) into the hash map
    put_Int2PtrHashMap(&path_data->nodes_map, from->index, NULL_NDFSM_ARC);

#ifdef NDFSM_NODE_WEIGHING
    if(find_nonfinished_node)
    {
        path_data->nodes_queue = create_PQueue(NDFSM_NODES_ALLOC_BY, false);
        assertion(path_data->nodes_queue != NULL, "init_PathFinderData: nodes_queue == NULL");

        from->distance = 0;

        insert_PQueue(path_data->nodes_queue, from->weight, (void*)0);
    }
#endif
}

static void destroy_PathFinderData(PathFinderData *path_data, int destroy_type)
{
    destroy_ArrayList(&path_data->nodes);

    if(destroy_type)
        { destroy_Int2PtrHashMap(&path_data->nodes_map); }
    else
        { reinit_Int2PtrHashMap(&path_data->nodes_map); }

#ifdef NDFSM_NODE_WEIGHING
    if(path_data->target_node == NULL)
        { destroy_PQueue(path_data->nodes_queue); }
#endif
}

/******************************************************************************/
/**                                NDFSM Data                                **/
/******************************************************************************/

/** @typedef NDFSM_Data
 * Internal representation of a graph (state of traversal).
 */
typedef struct NDFSM_Data
{
    /// Pointer to graph information.
    TE_GraphInfo *graph_info;
    /// List of nodes.
    ArrayList nodes;
    /// HashMap node_id -> index.
    HashMap *hash_map;
    /// Path finder data.
    PathFinderData path_data;
    /// Path to a nonfinished node (in reverse order).
    ArrayList path;
    /// Index of the last arc in the path (usually it is zero).
    int last_arc_index;
    /// Try nondeterministic arcs if there are no others.
    int nondeterm;
    /// Randomize choice of an untraversed arc.
    int randomize;
    /// Flag controlling reinitialization of the path finder's hash map.
    int initialize;

#if defined(NDFSM_NODE_WEIGHING) && !defined(NDFSM_OPTIMISTIC_WEIGHING)
    int path_num;
    int path_sum;
#endif
} NDFSM_Data;

static void init_NDFSM_Data(NDFSM_Data *data, TE_GraphInfo *info)
{
    assertion(data != NULL, "init_NDFSM_Data: data == NULL");
    assertion(info != NULL, "init_NDFSM_Data: info == NULL");

    data->graph_info = info;

    init_ArrayList(&data->nodes, sizeof(NDFSM_Node), NDFSM_NODES_ALLOC_BY, (Destructor)destroy_NDFSM_Node);
    init_ArrayList(&data->path,  sizeof(NDFSM_Arc*), NDFSM_PATH_ALLOC_BY,  NULL);

    data->last_arc_index = 0;

    data->hash_map = create_HashMap(&type_String, &type_Integer, NDFSM_NODES_HASH_SIZE, (HashFunc)hash_String);

    data->nondeterm  = 0;
    data->randomize  = 0;
    data->initialize = 1;

#if defined(NDFSM_NODE_WEIGHING) && !defined(NDFSM_OPTIMISTIC_WEIGHING)
    data->path_num = 0;
    data->path_sum = 0;
#endif
}

static void processParams_NDFSM_Data(NDFSM_Data *data, int argc, char **argv)
{
    int i;

    assertion(data != NULL,              "processParams_NDFSM_Data: data == NULL");
    assertion(argc >= 0,                 "processParams_NDFSM_Data: argc < 0");
    assertion(argv != NULL || argc == 0, "processParams_NDFSM_Data: argv == NULL && argc > 0");

    for(i = 0; i < argc; i++)
    {
        if(!strcmp("--cpu", argv[i]))
        {
            assertion(++i < argc, "processParams_NDFSM_Data: bad parameter: after -cpu should be an integer CPU identifier");
            data->graph_info->cpu_id = atoi(argv[i]);
        }
        else if(!strcmp("--nondeterministic", argv[i]))
        {
#ifdef NDFSM_ARC_GROUPING
            data->nondeterm = 1;
#else
            assertion(false, "processParams_NDFSM_Data: --nondeterministic mode is not supported");
#endif
        }
        else if(!strcmp("--randomize", argv[i]))
        {
            data->randomize = 1;
        }
    }
}

static void destroy_NDFSM_Data(NDFSM_Data *data)
{
    assertion(data != NULL, "destroy_NDFSM_Data: data == NULL");

    destroy_ArrayList(&data->nodes);
    destroy_ArrayList(&data->path);

    destroy(data->hash_map);

    // The path finder's hash map has not been destroyed
    if(!data->initialize)
        { destroy_Int2PtrHashMap(&data->path_data.nodes_map); }
}

static bool nonfinished_node_searcher(void *element, void *key)
{
    NDFSM_Node *node = (NDFSM_Node*)element;

    return node->new_arcs.size > 0;
}

static NDFSM_Node* findNonfinishedNode_NDFSM_Data(NDFSM_Data *data)
{
    return (NDFSM_Node*)find_ArrayList(&data->nodes, nonfinished_node_searcher, NULL);
}

static bool arc_finder_by_stimulus_id(void *element, void *key)
{
    NDFSM_Arc *arc = (NDFSM_Arc*)element;
    TE_ArcId *arc_id = (TE_ArcId*)key;

    return equal_StimulusId(arc->stimulus_id, arc_id->stimulus_id);
}

static bool arc_finder_by_stimulus_id_and_end_node_id(void *element, void *key)
{
    NDFSM_Node *end_node;
    NDFSM_Arc *arc = (NDFSM_Arc*)element;
    TE_ArcId *arc_id = (TE_ArcId*)key;
#ifdef NDFSM_ARC_GROUPING
    int i;
#endif

    if(!equal_StimulusId(arc->stimulus_id, arc_id->stimulus_id))
        { return false; }

    if(arc_id->end_node_id == NULL)
        { return true; }

#ifdef NDFSM_ARC_GROUPING
    for(i = 0; i < arc->end_nodes.size; i++)
    {
        end_node = GET_END_NODE_I(arc, i);
#else
        end_node = GET_END_NODE(arc);
#endif
        if(equals(r(end_node->node_id), r(arc_id->end_node_id)))
            { return true; }
#ifdef NDFSM_ARC_GROUPING
    }
#endif

    return false;
}

static NDFSM_Node* findNode_NDFSM_Data(NDFSM_Data *data, TE_NodeId *node_id)
{
    Integer *index;
    NDFSM_Node *node;

    // Look up into the hash map
    if((index = get_HashMap(r(data->hash_map), r(node_id->id))) == NULL)
        { return NULL; }

    node = (NDFSM_Node*)get_ArrayList(&data->nodes, *index);
    destroy(index);

    return node;
}

static NDFSM_Arc* findArcByStimulusId_ArcList(ArrayList *array, TE_ArcId *arc_id)
{
    return (NDFSM_Arc*)find_ArrayList(array, arc_finder_by_stimulus_id, arc_id);
}

static NDFSM_Arc* findArcByStimulusIdAndEndNodeId_ArcList(ArrayList *array,
    TE_ArcId *arc_id, TE_NodeId *end_node_id)
{
    NDFSM_Arc *arc;
    TE_NodeId *node_id;

    node_id = arc_id->end_node_id; arc_id->end_node_id = end_node_id;
    arc = (NDFSM_Arc*)find_ArrayList(array, arc_finder_by_stimulus_id_and_end_node_id, arc_id);
    arc_id->end_node_id = node_id;

    return arc;
}

static NDFSM_Node* registerNode_NDFSM_Data(NDFSM_Data *data, TE_UpdateType type, TE_NodeId *node_id)
{
    int index = data->nodes.size;
    NDFSM_Node *node = findNode_NDFSM_Data(data, node_id);

    if(node != NULL)
        { return node; }

    node = (NDFSM_Node*)append_ArrayList(&data->nodes);
    init_NDFSM_Node(node, index, node_id);

    // Update the hash map
    put_HashMap(r(data->hash_map), r(node_id->id), create_Integer(index));

    // Change statistics
    data->graph_info->nodes_traversed++;
    if(type == TE_LOCAL)
        { data->graph_info->nodes_traversed_locally++; }

    return node;
}

static NDFSM_Arc* registerNewArc_NDFSM_Data(NDFSM_Data *data, TE_UpdateType type, int index,
    TE_StimulusId stimulus_id, NDFSM_Node *start_node)
{
    NDFSM_Arc *creating_arc = NULL;

    creating_arc = (NDFSM_Arc*)append_ArrayList(&start_node->new_arcs);
    init_NDFSM_Arc(creating_arc, index, 0, stimulus_id, start_node, NULL);

    // Change statistics
    data->graph_info->arcs_known++;

    return creating_arc;
}

#ifdef NDFSM_ARC_GROUPING
static void change_arcs_order(NDFSM_Node *start_node, NDFSM_Arc *creating_arc)
{
    int i, j;

    i = index_ArrayList(&start_node->old_arcs, creating_arc);
    for(j = i + 1; j < start_node->old_arcs.size; j++)
    {
        NDFSM_Arc *arc = (NDFSM_Arc*)get_ArrayList(&start_node->old_arcs, j);

        if(arc->end_nodes.size >= creating_arc->end_nodes.size)
            { break; }
    }

    if(--j > i)
        { exchange_ArrayList(&start_node->old_arcs, i, j); }
}
#endif

static NDFSM_Arc* registerOldArc_NDFSM_Data(NDFSM_Data *data, TE_UpdateType type, int index,
    int flags, TE_StimulusId stimulus_id, NDFSM_Node *start_node, NDFSM_Node *end_node, NDFSM_Arc *creating_arc)
{
    // Loops are not added, but taken into account in the statistics
    if(start_node != end_node)
    {
#ifdef NDFSM_ARC_GROUPING
        // if there is an arc labeled with the same stimulus
        if(creating_arc != NULL)
        {
            if(creating_arc->end_nodes.size < NDFSM_MAX_ARC_GROUP_SIZE)
            {
                NDFSM_Node **node_holder = (NDFSM_Node**)append_ArrayList(&creating_arc->end_nodes);
               *node_holder = end_node;

                // Sort the outgoing arcs according to the nondeterminism degree
                if(data->nondeterm)
                    { change_arcs_order(start_node, creating_arc); }
            }
            else
            {
                // If there are too many arcs labeled with the same stimulus,
                // the stimulus is considered to be chaotic and not used further
                creating_arc->flags |= NDFSM_FLAG_CHAOTIC;
            }
        }
        else
        {
#endif
            creating_arc = (NDFSM_Arc*)append_ArrayList(&start_node->old_arcs);
            init_NDFSM_Arc(creating_arc, index, flags, stimulus_id, start_node, end_node);
#ifdef NDFSM_ARC_GROUPING
        }
#endif

    }
    else
    {
        // Increment number of loops
        data->graph_info->loops_traversed++;
    }

    // Change statistics
    data->graph_info->arcs_traversed++;
    if((flags & NDFSM_FLAG_NDETERM) != 0)
    {
        data->graph_info->arcs_known++;
        data->graph_info->arcs_nondeterministic++;
    }
    if(type == TE_LOCAL)
    {
        data->graph_info->arcs_traversed_locally++;
    }

    return creating_arc;
}

/******************************************************************************/
/**                             NDFSM Path Finder                            **/
/******************************************************************************/

static bool arc_analyser(void *element, void *key)
{
#ifdef NDFSM_ARC_GROUPING
    int i;
#endif
    NDFSM_Node *end_node;
    NDFSM_Arc *tmp;
    NDFSM_Arc *arc = (NDFSM_Arc*)element;
    PathFinderData *path_data = (PathFinderData*)key;
    NDFSM_Node **node_holder;
    bool find_nonfinished_node = (path_data->target_node == NULL);

    // Do not use chaotic arcs for finding a path
    if((arc->flags & NDFSM_FLAG_CHAOTIC) != 0)
        { return false; }

    // Do not use nondeterministic arcs if --nondeterministic is not set
    if(!path_data->nondeterm && (arc->flags & NDFSM_FLAG_NDETERM) != 0)
        { return false; }

#ifdef NDFSM_ARC_GROUPING
    // Calculate degree nondeterminism
    tmp = (NDFSM_Arc*)get_Int2PtrHashMap(&path_data->nodes_map, arc->start_node->index);

    arc->ndegree = (tmp == NULL_NDFSM_ARC ? 0 : tmp->ndegree) +
        ((arc->flags & NDFSM_FLAG_NDETERM) != 0 ? 1 : 0);

    for(i = 0; i < arc->end_nodes.size; i++)
    {
        end_node = GET_END_NODE_I(arc, i);
#else
        end_node = GET_END_NODE(arc);
#endif
        assertion(end_node != NULL, "arc_analyser: end_node == NULL");

        // The end node of the arc is a satisfactory one
        if(path_data->node_finder(end_node, path_data->target_node))
            { return true; }

        // The pair (end_node -> arc) is put into the hash map and the key is new
        if((tmp = (NDFSM_Arc*)put_Int2PtrHashMap(&path_data->nodes_map, end_node->index, arc)) == arc)
        {
#ifdef NDFSM_NODE_WEIGHING
            if(find_nonfinished_node)
            {
                end_node->distance = arc->start_node->distance + 1;

                insert_PQueue(path_data->nodes_queue, (end_node->weight + end_node->distance),
                    (void*)path_data->nodes.size);
            }
#endif
            // Append the end node into the nodes list
            node_holder = (NDFSM_Node**)append_ArrayList(&path_data->nodes);
           *node_holder = end_node;
        }
#ifdef NDFSM_ARC_GROUPING
        // The end_node already exists
        else if(path_data->nondeterm)
        {
            assertion(tmp != NULL, "arc_analyser: tmp == NULL");

            // If the new arc provides less degree of nondeterminism
            if(tmp != NULL_NDFSM_ARC && arc->ndegree < tmp->ndegree)
            {
                // Override the old arc with the new one
                override_Int2PtrHashMap(&path_data->nodes_map, end_node->index, arc);
            }
        }
    }
#endif

    return false;
}

static bool nonfinished_node_finder(NDFSM_Node *node, NDFSM_Node *target_node)
{
    return node->new_arcs.size > 0;
}

static bool target_node_finder(NDFSM_Node *node, NDFSM_Node *target_node)
{
    return equals(r(node->node_id), r(target_node->node_id));
}

static bool findPath_NDFSM_Data(NDFSM_Data *data, NDFSM_Node *from, NDFSM_Node *to,
    NodeFinder node_finder)
{
    int i, hash_size, index;
    NDFSM_Arc **arc_holder;
    NDFSM_Node *node, *end_node;
    NDFSM_Arc  *arc;
    bool verdict = false;
    bool find_nonfinished_node = (to == NULL);

    // Clear the path to be returned
    clear_ArrayList(&data->path);
    data->last_arc_index = 0;

#if defined(NDFSM_NODE_WEIGHING) && !defined(NDFSM_OPTIMISTIC_WEIGHING)
    if(data->path_num == NDFSM_AVG_PATH_SAMPLE)
    {
        data->path_sum /= data->path_num;
        data->path_num = 1;
    }

    data->path_num++;
    data->path_sum++;
#endif

    // The current node is a satisfactory one
    if(node_finder(from, to))
        { return true; }

    // Calculate size for path finder hash map as min(number_of_nodes, max_size)
    hash_size = data->nodes.size < NDFSM_PATH_FINDER_MAX_HASH_SIZE ?
        data->nodes.size : NDFSM_PATH_FINDER_MAX_HASH_SIZE;

    // Initialize the path finder
    init_PathFinderData(&data->path_data, from, to, node_finder, data->nondeterm,
        hash_size, NDFSM_PATH_FINDER_MIN_HASH_SIZE, data->initialize);

    // In the beginning the list of nodes contains the current node only
    for(i = 0; i < data->path_data.nodes.size; i++)
    {
#ifdef NDFSM_NODE_WEIGHING
        index = (find_nonfinished_node ? (int)extractMin_PQueue(data->path_data.nodes_queue) : i);
#else
        index = i;
#endif
        node = *(NDFSM_Node**)get_ArrayList(&data->path_data.nodes, index);
        arc  =  (NDFSM_Arc*)find_ArrayList(&node->old_arcs, arc_analyser, &data->path_data);

#ifdef NDFSM_NODE_WEIGHING
        if(find_nonfinished_node)
            { setWeight_NDFSM_Node(node); }
#endif

        // If the arc leads to a satisfactory node
        if(arc != NULL)
        {
            end_node = GET_END_NODE(arc);

#ifdef NDFSM_NODE_WEIGHING
            if(find_nonfinished_node && end_node->new_arcs.size > 1)
                { end_node->weight = NDFSM_UNKNOWN_WEIGHT(data); }
#endif
            // Construct a path in reverse order
            for(;;)
            {
                // Append the arc into the path
                arc_holder = (NDFSM_Arc**)append_ArrayList(&data->path);
               *arc_holder = arc;

#ifdef NDFSM_NODE_WEIGHING
                if(find_nonfinished_node)
                    { setWeight_NDFSM_Node(arc->start_node); }
#endif
                arc = (NDFSM_Arc*)get_Int2PtrHashMap(&data->path_data.nodes_map, arc->start_node->index);

                // The path has been successfully constructed
                if(arc == NULL_NDFSM_ARC)
                    { break; }

                // Consider the fragment ending with the first nondeterministic arc
                if((arc->flags & NDFSM_FLAG_NDETERM) != 0)
                   { data->last_arc_index = data->path.size - 1; }
            }

            verdict = true;
            break;
        }
    }

#if defined(NDFSM_NODE_WEIGHING) && !defined(NDFSM_OPTIMISTIC_WEIGHING)
    data->path_sum += data->path.size;
#endif

    if(data->initialize && hash_size >= NDFSM_PATH_FINDER_MAX_HASH_SIZE)
        { data->initialize = 0; }

    // Destroy the path finder
    destroy_PathFinderData(&data->path_data, data->initialize); 

    return verdict;
}

static NDFSM_Arc* chooseNewArc_NDFSM_Data(NDFSM_Data *data, NDFSM_Node *node)
{
    int i = node->new_arcs.size;
    assertion(i > 0, "chooseNewArc_NDFSM_Data: no new arcs");

    i = (data->randomize ? rand() & 0x7fffffff: data->graph_info->cpu_id) % i;

    return (NDFSM_Arc*)get_ArrayList(&node->new_arcs, i);
}

/******************************************************************************/
/**                              NDFSM Interface                             **/
/******************************************************************************/

static void init_NDFSM(TE_Graph *graph, int argc, char **argv)
{
    NDFSM_Data *data = (NDFSM_Data*)malloc(sizeof(NDFSM_Data));
    assertion(data != NULL, "init_NDFSM: out of memory");

    init_NDFSM_Data(data, &graph->info);
    processParams_NDFSM_Data(data, argc, argv);

    graph->data = data;
}

static void config_NDFSM(TE_Graph *graph, const char *attr_name, const char *attr_value)
{
    NDFSM_Data *data = (NDFSM_Data*)graph->data;

    if(!strcmp(TE_ATTR_CPU_ID, attr_name))
        { data->graph_info->cpu_id = atoi(attr_value); return; }

    assertion(false, "config_NDFSM: unknown attribute");
}

static TE_UpdateStatus start_NDFSM(TE_Graph *graph, TE_NodeId *node_id)
{
    NDFSM_Node *node;
    NDFSM_Data *data = (NDFSM_Data*)graph->data;
    TE_UpdateStatus status = TE_OLD_ALL;

    if((node = findNode_NDFSM_Data(data, node_id)) == NULL)
        { status = TE_NEW_NODE; }

    registerNode_NDFSM_Data(data, TE_LOCAL, node_id);

    return status;
}

static TE_UpdateStatus update_NDFSM(TE_Graph *graph, TE_UpdateType type,
    TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
    int i;
    NDFSM_Arc  *arc;
    NDFSM_Node *start_node;
    NDFSM_Node *end_node;
    NDFSM_Node *arc_end_node;
    NDFSM_Data *data = (NDFSM_Data*)graph->data;
    TE_NodeId  *end_node_id;
    TE_UpdateStatus status = TE_OLD_ALL;

    assertion(arc_id != NULL, "update_NDFSM: arc_id == NULL");

    start_node = findNode_NDFSM_Data(data, arc_id->start_node_id);

#ifdef DEBUG_NETFSM
    if(start_node == NULL)
    {
        printf("update_NDFSM: unconnected subgraph: update=%s, start_node_id=%s\n",
            type == TE_REMOTE ? "remote" : "local", NODE_ID_STRING(arc_id->start_node_id));
    }
#endif

    assertion(start_node != NULL, "update_NDFSM: unconnected subgraph");

    end_node_id = (new_end_node_id != NULL ? new_end_node_id : arc_id->end_node_id);

    // Add the untraversed arc
    if(end_node_id == NULL)
    {
        assertion(start_node->old_arcs.size == 0, "update_NDFSM: new arcs are added first");
        assertion(start_node->new_arcs.size < start_node->node_id->deg, "update_NDFSM: an odd new arc");

        i = data->graph_info->arcs_known;
        registerNewArc_NDFSM_Data(data, type, i, arc_id->stimulus_id, start_node);
        status |= TE_NEW_ARC;

#ifdef NDFSM_NODE_WEIGHING
        start_node->weight = NDFSM_INITIAL_WEIGHT(start_node);
#endif
    }
    // Add the traversed arc
    else
    {
        if((end_node = findNode_NDFSM_Data(data, end_node_id)) == NULL)
        {
            // Add the new node
            end_node = registerNode_NDFSM_Data(data, type, end_node_id);
            status |= TE_NEW_NODE;
        }
        else
        {
            // Set degree of the node
            end_node_id->deg = end_node->node_id->deg;
        }

        // There is a traversed arc labeled with the same stimulus
        if((arc = findArcByStimulusId_ArcList(&start_node->old_arcs, arc_id)) != NULL)
        {
            // The stimulus may cause nondeterminism
            if((arc->flags & NDFSM_FLAG_NDETERM) != 0)
            {
                // The updated arc has a new end node
                if((arc = findArcByStimulusIdAndEndNodeId_ArcList(&start_node->old_arcs, arc_id, end_node_id)) == NULL)
                {
                    i = data->graph_info->arcs_known;
                    registerOldArc_NDFSM_Data(data, type, i, NDFSM_FLAG_NDETERM, arc_id->stimulus_id, start_node, end_node, arc);
                    status |= TE_NEW_ARC;
                }
            }
            // The stimulus has not caused nondeterminism so far
            else
            {
                arc_end_node = GET_END_NODE(arc);

                // However, it is nondeterministic
                if(!equals(r(arc_end_node->node_id), r(end_node_id)))
                {
                    arc->flags |= NDFSM_FLAG_NDETERM;
                    data->graph_info->arcs_nondeterministic++;

                    i = data->graph_info->arcs_known;
                    registerOldArc_NDFSM_Data(data, type, i, NDFSM_FLAG_NDETERM, arc_id->stimulus_id, start_node, end_node, arc);
                    status |= TE_NEW_ARC;
                }
            }
        }
        // There are no traversed arcs labeled with the same stimulus
        else
        {
            // Remove the arc from the list of untraversed arcs if it has been found
            if((arc = findArcByStimulusId_ArcList(&start_node->new_arcs, arc_id)) != NULL)
            {
                i = data->graph_info->arcs_known;
                registerOldArc_NDFSM_Data(data, type, i, 0, arc_id->stimulus_id, start_node, end_node, NULL);
                status |= TE_NEW_ARC;

                remove_ArrayList(&start_node->new_arcs, arc);

                // The node has finished
                if(start_node->new_arcs.size == 0)
                {
                    // Destroy the list of untraversed arcs to deallocate memory
                    destroy_ArrayList(&start_node->new_arcs);

                    // Change statistics
                    data->graph_info->nodes_finished++;
#ifdef NDFSM_NODE_WEIGHING
                    start_node->weight = NDFSM_UNKNOWN_WEIGHT(data);
#endif
                }
            }
            else
            {
                // If the untraversed arc does not exist, it means that it has been
                // already traversed but not added to the list of the traversed arcs.
                // It is possible if and only if the arc was a loop.

                // If the added arc is not a loop, then there is nondeterminism
                if(start_node != end_node)
                {
                    // The previous loop is nondeterministic
                    data->graph_info->arcs_nondeterministic++;

                    i = data->graph_info->arcs_known;
                    registerOldArc_NDFSM_Data(data, type, i, NDFSM_FLAG_NDETERM, arc_id->stimulus_id, start_node, end_node, NULL);
                    status |= TE_NEW_ARC;
                }
            }
        }
    }

    return status;
}

static TE_ArcList* getPath_NDFSM_Data(NDFSM_Data *data)
{
    int i;
    NDFSM_Arc *arc;
    TE_ArcList *path = create_List(&type_TE_ArcId);

    for(i = data->path.size - 1; i >= data->last_arc_index; i--)
    {
        arc = *((NDFSM_Arc**)get_ArrayList(&data->path, i));
        append_List(r(path), createArcId_NDFSM_Arc(arc));
    }

    return path;
}

static TE_ArcList* getPath_NDFSM(TE_Graph *graph, TE_NodeId *node_id)
{
    TE_ArcList *path;
    NDFSM_Node *node;
    NDFSM_Arc  *arc;
    NDFSM_Arc  *last_arc = NULL;
    NDFSM_Data *data = (NDFSM_Data*)graph->data;

    // Traversal has been completed
    if(data->graph_info->arcs_traversed >= data->graph_info->arcs_known)
        { graph->error = NDFSM_ERRNO_NO_ERRORS; return NULL; }

    // Get the current node
    node = findNode_NDFSM_Data(data, node_id);
    assertion(node != NULL, "getPath_NDFSM: node == NULL");

    // Unconnected graph
    if(!findPath_NDFSM_Data(data, node, NULL, nonfinished_node_finder))
        { graph->error = NDFSM_ERRNO_UNCONNECT; return NULL; }

    // Get the last node of the path
    if(data->path.size > 0)
    {
        // If the arc is nondeterministic, we take the first variant
        last_arc = *((NDFSM_Arc**)get_ArrayList(&data->path, data->last_arc_index));
        node = GET_END_NODE(last_arc);
    }

    // Convert the path into the external format
    path = getPath_NDFSM_Data(data);
 
    // Full path should be ended with an untraversed arc
    if(data->last_arc_index == 0 && (last_arc == NULL || (last_arc->flags & NDFSM_FLAG_NDETERM) == 0))
    {
        // Choose a new arc according to the CPU identifier
        arc = chooseNewArc_NDFSM_Data(data, node);
        append_List(r(path), createArcId_NDFSM_Arc(arc));
    }

    return path;
}

static TE_ArcList* reachNode_NDFSM(TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
{
    TE_ArcList *path;
    NDFSM_Node *start_node;
    NDFSM_Node *target_node;
    NDFSM_Data *data = (NDFSM_Data*)graph->data;

    // Get the initial node
    start_node = (NDFSM_Node*)get_ArrayList(&data->nodes, 0);
    assertion(start_node != NULL, "reachNode_NDFSM: start_node == NULL");

    // Get the target node
    target_node = findNode_NDFSM_Data(data, node_id);
    assertion(target_node != NULL, "reachNode_NDFSM: target_node == NULL");

    // Try to find the path from the initial node to the target one
    if(!findPath_NDFSM_Data(data, start_node, target_node, target_node_finder))
        { return NULL; }

    // Convert the path into the external format
    path = getPath_NDFSM_Data(data);

    // TODO: there can be non deterministic arc

    if(arc_id != NULL)
        { append_List(r(path), r(arc_id)); }

    return path;
}

static String* getUnconnectedGraphDiagnostic_NDFSM_Data(NDFSM_Data *data)
{
    if(data != NULL)
    {
        const char *error_reason = data->graph_info->arcs_nondeterministic == 0 ?
            "incomplete set of actions or too strong preconditions" :
#ifdef NDFSM_ARC_GROUPING
            (data->nondeterm ? "chaotic behavior" : "nondeterminism with no '--nondeterministic' set");
#else
            "nondeterminism";
#endif

        NDFSM_Node *nonfinished_node = findNonfinishedNode_NDFSM_Data(data);

        return format_String("Unconnected graph: nonfinished state %s could not be reached (%s)",
            NODE_STRING(nonfinished_node), error_reason);
    }

    // Graph has been already destroyed
    return create_String("Unconnected graph");
}

static String* getError_NDFSM(TE_Graph *graph)
{
    NDFSM_Data *data = (NDFSM_Data*)graph->data;

    switch(graph->error)
    {
    case NDFSM_ERRNO_NO_ERRORS:
        { return create_String("None"); }
    case NDFSM_ERRNO_UNCONNECT:
        { return getUnconnectedGraphDiagnostic_NDFSM_Data(data); }
    default:
        { return create_String("Unknown error"); }
    }
}

static void finish_NDFSM(TE_Graph *graph)
{
    NDFSM_Data *data = (NDFSM_Data*)graph->data;

    if(data != NULL)
    {
        destroy_NDFSM_Data(data);
        free(data);
        graph->data = NULL;
    }
}

static void toFile_NDFSM_Arc(FILE *file, NDFSM_Arc *arc)
{
    NDFSM_Node *end_node;

#ifdef NDFSM_ARC_GROUPING
    int i;
    for(i = 0; i < arc->end_nodes.size; i++)
    {
        end_node = GET_END_NODE_I(arc, i);
#else
        end_node = GET_END_NODE(arc);
#endif

        fprintf(file, "  <transition>\n");
#ifdef NDFSM_NODE_WEIGHING
        fprintf(file, "    <start_node weight=%d>%s</start_node>\n",
            arc->start_node->weight, NODE_STRING(arc->start_node));
#else
        fprintf(file, "    <start_node>%s</start_node>\n", NODE_STRING(arc->start_node));
#endif
        fprintf(file, "      <stimulus>\n");
        fprintf(file, "        <index>%d</index>\n", arc->stimulus_id.index);
        fprintf(file, "      </stimulus>\n");
        fprintf(file, "    <end_node>%s</end_node>\n", NODE_STRING(end_node));
        fprintf(file, "  </transition>\n");

#ifdef NDFSM_ARC_GROUPING
    }
#endif
}

static bool toFile_NDFSM(TE_Graph *graph, const char *filename)
{
    int i, j;
    FILE *file;
    NDFSM_Arc  *arc;
    NDFSM_Node *node;
    NDFSM_Data *data = (NDFSM_Data*)graph->data;

    if((file = fopen(filename, "w")) == NULL)
        { return false; }

    fprintf(file, "<graph>\n");

    for(i = 0; i < data->nodes.size; i++)
    {
        node = (NDFSM_Node*)get_ArrayList(&data->nodes, i);

        for(j = 0; j < node->new_arcs.size; j++)
        {
            arc = (NDFSM_Arc*)get_ArrayList(&node->new_arcs, j);
            toFile_NDFSM_Arc(file, arc);
        }

        for(j = 0; j < node->old_arcs.size; j++)
        {
            arc = (NDFSM_Arc*)get_ArrayList(&node->old_arcs, j);
            toFile_NDFSM_Arc(file, arc);
        }
    }

    fprintf(file, "</graph>\n");
    fclose(file);

    return true;
}

/******************************************************************************/
/**                               NDFSM Engine                               **/
/******************************************************************************/

TestEngine TE_NDFSM =
{
    /* .name      = */ "netndfsm",
    /* .init      = */ init_NDFSM,
    /* .config    = */ config_NDFSM,
    /* .start     = */ start_NDFSM,
    /* .update    = */ update_NDFSM,
    /* .getPath   = */ getPath_NDFSM,
    /* .reachNode = */ reachNode_NDFSM,
    /* .getError  = */ getError_NDFSM,
    /* .finish    = */ finish_NDFSM,
    /* .toFile    = */ toFile_NDFSM,
    /* .fromFile  = */ NULL,
    /* .sysdata   = */ { NULL }
};
